import Koa from 'koa';
import Router from 'koa-router';

import {get_setting, SettingConfig} from './SettingConfig';
import {connectDB} from './database';
import {DiffManager, createDiffRouter} from './diff/mod';

import { createReadStream, readFileSync } from 'fs';
import getContentRouter from './route/contents';
import { createKnexDocumentAccessor, createKnexTagController, createKnexUserController } from './db/mod';
import bodyparser from 'koa-bodyparser';
import {error_handler} from './route/error_handler';
import {createUserMiddleWare, createLoginRouter, isAdminFirst, getAdmin} from './login';

import {createInterface as createReadlineInterface} from 'readline';
import { DocumentAccessor, UserAccessor, TagAccessor } from './model/mod';
import { createComicWatcher } from './diff/watcher/comic_watcher';
import { getTagRounter } from './route/tags';

class ServerApplication{
    readonly userController: UserAccessor;
    readonly documentController: DocumentAccessor;
    readonly tagController: TagAccessor;
    readonly diffManger: DiffManager;
    readonly app: Koa;
    private index_html:Buffer;
    private constructor(controller:{
            userController: UserAccessor,
            documentController:DocumentAccessor,
            tagController: TagAccessor}){
        this.userController = controller.userController;
        this.documentController = controller.documentController;
        this.tagController = controller.tagController;
        
        this.diffManger = new DiffManager(this.documentController);
        this.app = new Koa();
        this.index_html = readFileSync("index.html");
    }
    private async setup(){
        const setting = get_setting();
        const app = this.app;

        if(setting.cli){
            const userAdmin = await getAdmin(this.userController);
            if(await isAdminFirst(userAdmin)){
                const rl = createReadlineInterface({
                    input:process.stdin,
                    output:process.stdout
                });
                const pw = await new Promise((res:(data:string)=>void,err)=>{
                    rl.question("put admin password :",(data)=>{
                        res(data);
                    });
                });
                rl.close();
                userAdmin.reset_password(pw);
            }
        }
        app.use(bodyparser());
        app.use(error_handler);
        app.use(createUserMiddleWare(this.userController));
        
        let diff_router = createDiffRouter(this.diffManger);
        this.diffManger.register("comic",createComicWatcher());
        let router = new Router();
        
        router.use('/api/diff',diff_router.routes());
        router.use('/api/diff',diff_router.allowedMethods());

        this.serve_index(router);
        this.serve_static_file(router);
    
        const content_router = getContentRouter(this.documentController);
        router.use('/api/doc',content_router.routes());
        router.use('/api/doc',content_router.allowedMethods());

        const tags_router = getTagRounter(this.tagController);
        router.use("/api/tags",tags_router.routes());
        router.use("/api/tags",tags_router.allowedMethods());

        const login_router = createLoginRouter(this.userController);
        router.use('/user',login_router.routes());
        router.use('/user',login_router.allowedMethods());
        
        
        if(setting.mode == "development"){
        let mm_count = 0;
        app.use(async (ctx,next)=>{
            console.log(`==========================${mm_count++}`);
            const ip = (ctx.get("X-Real-IP")) ?? ctx.ip;
            const fromClient = ctx.state['user'].username === "" ? ip : ctx.state['user'].username;
            console.log(`${fromClient} : ${ctx.method} ${ctx.url}`);
            await next();
            //console.log(`404`);
        });}
        app.use(router.routes());
        app.use(router.allowedMethods());
    }
    private serve_index(router:Router){
        const serveindex = (url:string)=>{
            router.get(url, (ctx)=>{
                ctx.type = 'html'; ctx.body = this.index_html;
                const setting = get_setting();
                ctx.set('x-content-type-options','no-sniff');
                if(setting.mode === "development"){
                    ctx.set('cache-control','no-cache'); 
                }
                else{
                    ctx.set('cache-control','public, max-age=3600');
                }
            })
        }
        serveindex('/');
        serveindex('/doc/:rest(.*)');
        serveindex('/search');
        serveindex('/login');
        serveindex('/profile');
        serveindex('/difference');
        serveindex('/setting');
        serveindex('/tags');
    }
    private serve_static_file(router: Router){
        const static_file_server = (path:string,type:string) => {
            router.get('/'+path,async (ctx,next)=>{
                const setting = get_setting();
                ctx.type = type; ctx.body = createReadStream(path);
                ctx.set('x-content-type-options','no-sniff');
                if(setting.mode === "development"){
                    ctx.set('cache-control','no-cache'); 
                }
                else{
                    ctx.set('cache-control','public, max-age=3600');
                }
            })};
        const setting = get_setting();
        static_file_server('dist/bundle.css','css');
        static_file_server('dist/bundle.js','js');
        if(setting.mode === "development"){
            static_file_server('dist/bundle.js.map','text');        
            static_file_server('dist/bundle.css.map','text');
        }
    }
    start_server(){
        let setting = get_setting();
        console.log("start server");
        //todo : support https
        return this.app.listen(setting.port,setting.localmode ? "127.0.0.1" : "0.0.0.0");
    }
    static async createServer(){
        const setting = get_setting();
        let db = await connectDB();

        const app = new ServerApplication({
            userController:createKnexUserController(db),
            documentController: createKnexDocumentAccessor(db),
            tagController: createKnexTagController(db),
        });
        await app.setup();
        return app;
    }
}
//let Koa = require("koa");

export async function create_server(){
    return await ServerApplication.createServer();
}

export default {create_server};